# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest
import aniso8601

class TestInitFunctions(unittest.TestCase):
    def test_import(self):
        #Verify the function mappings
        self.assertEqual(aniso8601.parse_datetime,
                         aniso8601.time.parse_datetime)
        self.assertEqual(aniso8601.parse_time, aniso8601.time.parse_time)
        self.assertEqual(aniso8601.get_time_resolution,
                         aniso8601.time.get_time_resolution)
        self.assertEqual(aniso8601.get_datetime_resolution,
                         aniso8601.time.get_datetime_resolution)

        self.assertEqual(aniso8601.parse_date, aniso8601.date.parse_date)
        self.assertEqual(aniso8601.get_date_resolution,
                         aniso8601.date.get_date_resolution)

        self.assertEqual(aniso8601.parse_duration,
                         aniso8601.duration.parse_duration)
        self.assertEqual(aniso8601.get_duration_resolution,
                         aniso8601.duration.get_duration_resolution)

        self.assertEqual(aniso8601.parse_interval,
                         aniso8601.interval.parse_interval)
        self.assertEqual(aniso8601.parse_repeating_interval,
                         aniso8601.interval.parse_repeating_interval)
        self.assertEqual(aniso8601.get_interval_resolution,
                         aniso8601.interval.get_interval_resolution)
        self.assertEqual(aniso8601.get_repeating_interval_resolution,
                         aniso8601.interval.get_repeating_interval_resolution)



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.exceptions import ISOFormatError
from aniso8601.builders import TupleBuilder
from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.compat import is_string
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import DateResolution

def get_date_resolution(isodatestr):
    #Valid string formats are:
    #
    #Y[YYY]
    #YYYY-MM-DD
    #YYYYMMDD
    #YYYY-MM
    #YYYY-Www
    #YYYYWww
    #YYYY-Www-D
    #YYYYWwwD
    #YYYY-DDD
    #YYYYDDD
    isodatetuple = parse_date(isodatestr, builder=TupleBuilder)

    if isodatetuple.DDD is not None:
        #YYYY-DDD
        #YYYYDDD
        return DateResolution.Ordinal

    if isodatetuple.D is not None:
        #YYYY-Www-D
        #YYYYWwwD
        return DateResolution.Weekday

    if isodatetuple.Www is not None:
        #YYYY-Www
        #YYYYWww
        return DateResolution.Week

    if isodatetuple.DD is not None:
        #YYYY-MM-DD
        #YYYYMMDD
        return DateResolution.Day

    if isodatetuple.MM is not None:
        #YYYY-MM
        return DateResolution.Month

    #Y[YYY]
    return DateResolution.Year

def parse_date(isodatestr, builder=PythonTimeBuilder):
    #Given a string in any ISO 8601 date format, return a datetime.date
    #object that corresponds to the given date. Valid string formats are:
    #
    #Y[YYY]
    #YYYY-MM-DD
    #YYYYMMDD
    #YYYY-MM
    #YYYY-Www
    #YYYYWww
    #YYYY-Www-D
    #YYYYWwwD
    #YYYY-DDD
    #YYYYDDD
    if is_string(isodatestr) is False:
        raise ValueError('Date must be string.')

    if isodatestr.startswith('+') or isodatestr.startswith('-'):
        raise NotImplementedError('ISO 8601 extended year representation '
                                  'not supported.')

    if len(isodatestr) == 0 or isodatestr.count('-') > 2:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 date.'
                             .format(isodatestr))
    yearstr = None
    monthstr = None
    daystr = None
    weekstr = None
    weekdaystr = None
    ordinaldaystr = None

    if len(isodatestr) <= 4:
        #Y[YYY]
        yearstr = isodatestr
    elif 'W' in isodatestr:
        if len(isodatestr) == 10:
            #YYYY-Www-D
            yearstr = isodatestr[0:4]
            weekstr = isodatestr[6:8]
            weekdaystr = isodatestr[9]
        elif len(isodatestr) == 8:
            if '-' in isodatestr:
                #YYYY-Www
                yearstr = isodatestr[0:4]
                weekstr = isodatestr[6:]
            else:
                #YYYYWwwD
                yearstr = isodatestr[0:4]
                weekstr = isodatestr[5:7]
                weekdaystr = isodatestr[7]
        elif len(isodatestr) == 7:
            #YYYYWww
            yearstr = isodatestr[0:4]
            weekstr = isodatestr[5:]
    elif len(isodatestr) == 7:
        if '-' in isodatestr:
            #YYYY-MM
            yearstr = isodatestr[0:4]
            monthstr = isodatestr[5:]
        else:
            #YYYYDDD
            yearstr = isodatestr[0:4]
            ordinaldaystr = isodatestr[4:]
    elif len(isodatestr) == 8:
        if '-' in isodatestr:
            #YYYY-DDD
            yearstr = isodatestr[0:4]
            ordinaldaystr = isodatestr[5:]
        else:
            #YYYYMMDD
            yearstr = isodatestr[0:4]
            monthstr = isodatestr[4:6]
            daystr = isodatestr[6:]
    elif len(isodatestr) == 10:
        #YYYY-MM-DD
        yearstr = isodatestr[0:4]
        monthstr = isodatestr[5:7]
        daystr = isodatestr[8:]
    else:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 date.'
                             .format(isodatestr))

    hascomponent = False

    for componentstr in [yearstr, monthstr, daystr, weekstr, weekdaystr, ordinaldaystr]:
        if componentstr is not None:
            hascomponent = True

            if componentstr.isdigit() is False:
                raise ISOFormatError('"{0}" is not a valid ISO 8601 date.'
                                     .format(isodatestr))

    if hascomponent is False:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 date.'
                             .format(isodatestr))

    return builder.build_date(YYYY=yearstr, MM=monthstr, DD=daystr, Www=weekstr, D=weekdaystr, DDD=ordinaldaystr)



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.builders import TupleBuilder
from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.compat import is_string
from aniso8601.date import parse_date
from aniso8601.decimalfraction import normalize
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import TimeResolution
from aniso8601.timezone import parse_timezone

TIMEZONE_DELIMITERS = ['Z', '+', '-']

def get_time_resolution(isotimestr):
    #Valid time formats are:
    #
    #hh:mm:ss
    #hhmmss
    #hh:mm
    #hhmm
    #hh
    #hh:mm:ssZ
    #hhmmssZ
    #hh:mmZ
    #hhmmZ
    #hhZ
    #hh:mm:ss±hh:mm
    #hhmmss±hh:mm
    #hh:mm±hh:mm
    #hhmm±hh:mm
    #hh±hh:mm
    #hh:mm:ss±hhmm
    #hhmmss±hhmm
    #hh:mm±hhmm
    #hhmm±hhmm
    #hh±hhmm
    #hh:mm:ss±hh
    #hhmmss±hh
    #hh:mm±hh
    #hhmm±hh
    #hh±hh
    isotimetuple = parse_time(isotimestr, builder=TupleBuilder)

    return _get_time_resolution(isotimetuple)

def get_datetime_resolution(isodatetimestr, delimiter='T'):
    #<date>T<time>
    #
    #Time part cannot be omittted so return time resolution
    isotimetuple = parse_datetime(isodatetimestr, delimiter=delimiter, builder=TupleBuilder).time

    return _get_time_resolution(isotimetuple)

def _get_time_resolution(isotimetuple):
    if isotimetuple.ss is not None:
        return TimeResolution.Seconds

    if isotimetuple.mm is not None:
        return TimeResolution.Minutes

    return TimeResolution.Hours

def parse_time(isotimestr, builder=PythonTimeBuilder):
    #Given a string in any ISO 8601 time format, return a datetime.time object
    #that corresponds to the given time. Fixed offset tzdata will be included
    #if UTC offset is given in the input string. Valid time formats are:
    #
    #hh:mm:ss
    #hhmmss
    #hh:mm
    #hhmm
    #hh
    #hh:mm:ssZ
    #hhmmssZ
    #hh:mmZ
    #hhmmZ
    #hhZ
    #hh:mm:ss±hh:mm
    #hhmmss±hh:mm
    #hh:mm±hh:mm
    #hhmm±hh:mm
    #hh±hh:mm
    #hh:mm:ss±hhmm
    #hhmmss±hhmm
    #hh:mm±hhmm
    #hhmm±hhmm
    #hh±hhmm
    #hh:mm:ss±hh
    #hhmmss±hh
    #hh:mm±hh
    #hhmm±hh
    #hh±hh
    if is_string(isotimestr) is False:
        raise ValueError('Time must be string.')

    if len(isotimestr) == 0:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 time.'
                             .format(isotimestr))

    timestr = normalize(isotimestr)

    hourstr = None
    minutestr = None
    secondstr = None
    tzstr = None

    fractionalstr = None

    #Split out the timezone
    for delimiter in TIMEZONE_DELIMITERS:
        delimiteridx = timestr.find(delimiter)

        if delimiteridx != -1:
            tzstr = timestr[delimiteridx:]
            timestr = timestr[0:delimiteridx]

    #Split out the fractional component
    if timestr.find('.') != -1:
        timestr, fractionalstr = timestr.split('.', 1)

        if fractionalstr.isdigit() is False:
            raise ISOFormatError('"{0}" is not a valid ISO 8601 time.'
                                 .format(isotimestr))


    if len(timestr) == 2:
        #hh
        hourstr = timestr
    elif len(timestr) == 4 or len(timestr) == 5:
        #hh:mm
        #hhmm
        if timestr.count(':') == 1:
            hourstr, minutestr = timestr.split(':')
        else:
            hourstr = timestr[0:2]
            minutestr = timestr[2:]
    elif len(timestr) == 6 or len(timestr) == 8:
        #hh:mm:ss
        #hhmmss
        if timestr.count(':') == 2:
            hourstr, minutestr, secondstr = timestr.split(':')
        else:
            hourstr = timestr[0:2]
            minutestr = timestr[2:4]
            secondstr = timestr[4:]
    else:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 time.'
                             .format(isotimestr))

    for componentstr in [hourstr, minutestr, secondstr]:
        if componentstr is not None and componentstr.isdigit() is False:
            raise ISOFormatError('"{0}" is not a valid ISO 8601 time.'
                                 .format(isotimestr))

    if fractionalstr is not None:
        if secondstr is not None:
            secondstr = secondstr + '.' + fractionalstr
        elif minutestr is not None:
            minutestr = minutestr + '.' + fractionalstr
        else:
            hourstr = hourstr + '.' + fractionalstr

    if tzstr is None:
        tz = None
    else:
        tz = parse_timezone(tzstr, builder=TupleBuilder)

    return builder.build_time(hh=hourstr, mm=minutestr, ss=secondstr, tz=tz)

def parse_datetime(isodatetimestr, delimiter='T', builder=PythonTimeBuilder):
    #Given a string in ISO 8601 date time format, return a datetime.datetime
    #object that corresponds to the given date time.
    #By default, the ISO 8601 specified T delimiter is used to split the
    #date and time (<date>T<time>). Fixed offset tzdata will be included
    #if UTC offset is given in the input string.
    if is_string(isodatetimestr) is False:
        raise ValueError('Date time must be string.')

    if delimiter not in isodatetimestr:
        raise ISOFormatError('Delimiter "{0}" is not in combined date time '
                             'string "{1}".'
                             .format(delimiter, isodatetimestr))

    isodatestr, isotimestr = isodatetimestr.split(delimiter, 1)

    datepart = parse_date(isodatestr, builder=TupleBuilder)

    timepart = parse_time(isotimestr, builder=TupleBuilder)

    return builder.build_datetime(datepart, timepart)



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest
import aniso8601

from aniso8601.builders import (BaseTimeBuilder, DateTuple, DatetimeTuple,
                                DurationTuple, IntervalTuple,
                                RepeatingIntervalTuple, TimeTuple,
                                TimezoneTuple, TupleBuilder, cast)
from aniso8601.exceptions import (DayOutOfBoundsError, HoursOutOfBoundsError,
                                  ISOFormatError, LeapSecondError,
                                  MidnightBoundsError, MinutesOutOfBoundsError,
                                  MonthOutOfBoundsError,
                                  SecondsOutOfBoundsError,
                                  WeekOutOfBoundsError)
from aniso8601.tests.compat import mock

class LeapSecondSupportingTestBuilder(BaseTimeBuilder):
    LEAP_SECONDS_SUPPORTED = True

class TestBuilderFunctions(unittest.TestCase):
    def test_cast(self):
        self.assertEqual(cast('1', int), 1)
        self.assertEqual(cast('-2', int), -2)
        self.assertEqual(cast('3', float), float(3))
        self.assertEqual(cast('-4', float), float(-4))
        self.assertEqual(cast('5.6', float), 5.6)
        self.assertEqual(cast('-7.8', float), -7.8)

    def test_cast_exception(self):
        with self.assertRaises(ISOFormatError):
            cast('asdf', int)

        with self.assertRaises(ISOFormatError):
            cast('asdf', float)

    def test_cast_caughtexception(self):
        def tester(value):
            raise RuntimeError

        with self.assertRaises(ISOFormatError):
            cast('asdf', tester, caughtexceptions=(RuntimeError,))

    def test_cast_thrownexception(self):
        with self.assertRaises(RuntimeError):
            cast('asdf', int, thrownexception=RuntimeError)

class TestBaseTimeBuilder(unittest.TestCase):
    def test_build_date(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_date()

    def test_build_time(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_time()

    def test_build_datetime(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_datetime(None, None)

    def test_build_duration(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_duration()

    def test_build_interval(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_interval()

    def test_build_repeating_interval(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_repeating_interval()

    def test_build_timezone(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_timezone()

    def test_range_check_date(self):
        #Check the calendar for day ranges
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='0007', MM='02', DD='30')

        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='0007', DDD='366')

        with self.assertRaises(MonthOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='4333', MM='30', DD='30')

        #0 isn't a valid week number
        with self.assertRaises(WeekOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='2003', Www='00')

        #Week must not be larger than 53
        with self.assertRaises(WeekOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='2004', Www='54')

        #0 isn't a valid day number
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='2001', Www='02', D='0')

        #Day must not be larger than 7
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='2001', Www='02', D='8')

        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='1981', DDD='000')

        #Day must be 365, or 366, not larger
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='1234', DDD='000')

        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='1234', DDD='367')

        #https://bitbucket.org/nielsenb/aniso8601/issues/14/parsing-ordinal-dates-should-only-allow
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY='1981', DDD='366')

        #Make sure Nones pass through unmodified
        self.assertEqual(BaseTimeBuilder.range_check_date(rangedict={}), (None, None, None,
                                                                          None, None, None))

    def test_range_check_time(self):
        #Leap seconds not supported
        #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
        #https://bitbucket.org/nielsenb/aniso8601/issues/13/parsing-of-leap-second-gives-wildly
        with self.assertRaises(LeapSecondError):
            BaseTimeBuilder.range_check_time(hh='23', mm='59', ss='60')

        with self.assertRaises(SecondsOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh='00', mm='00', ss='60')

        with self.assertRaises(SecondsOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh='00', mm='00', ss='61')

        with self.assertRaises(MinutesOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh='00', mm='61')

        with self.assertRaises(MinutesOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh='00', mm='60')

        with self.assertRaises(MinutesOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh='00', mm='60.1')

        with self.assertRaises(HoursOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh='25')

        #Hour 24 can only represent midnight
        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh='24', mm='00', ss='01')

        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh='24', mm='00.1')

        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh='24', mm='01')

        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh='24.1')

        #Leap seconds not supported
        #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
        #https://bitbucket.org/nielsenb/aniso8601/issues/13/parsing-of-leap-second-gives-wildly
        with self.assertRaises(LeapSecondError):
            BaseTimeBuilder.range_check_time(hh='23', mm='59', ss='60')

        #Make sure Nones pass through unmodified
        self.assertEqual(BaseTimeBuilder.range_check_time(rangedict={}), (None, None, None, None))

    def test_range_check_time_leap_seconds_supported(self):
        self.assertEqual(LeapSecondSupportingTestBuilder.range_check_time(hh='23', mm='59', ss='60'), (23, 59, 60, None))

        with self.assertRaises(SecondsOutOfBoundsError):
            LeapSecondSupportingTestBuilder.range_check_time(hh='01', mm='02', ss='60')

    def test_range_check_duration(self):
        self.assertEqual(BaseTimeBuilder.range_check_duration(), (None, None, None,
                                                                  None, None, None,
                                                                  None))

        self.assertEqual(BaseTimeBuilder.range_check_duration(rangedict={}), (None, None, None,
                                                                              None, None, None,
                                                                              None))

    def test_range_check_repeating_interval(self):
        self.assertEqual(BaseTimeBuilder.range_check_repeating_interval(), (None, None, None))

        self.assertEqual(BaseTimeBuilder.range_check_repeating_interval(rangedict={}), (None, None, None))

    def test_range_check_timezone(self):
        self.assertEqual(BaseTimeBuilder.range_check_timezone(), (None, None, None,
                                                                  None, ''))

        self.assertEqual(BaseTimeBuilder.range_check_timezone(rangedict={}), (None, None, None,
                                                                              None, ''))

    def test_build_object(self):
        datetest = (DateTuple('1', '2', '3', '4', '5', '6'),
                    {'YYYY': '1', 'MM': '2', 'DD': '3',
                     'Www': '4', 'D': '5', 'DDD': '6'})

        timetest = (TimeTuple('1', '2', '3',
                              TimezoneTuple(False, False, '4', '5',
                                            'tz name')),
                    {'hh': '1', 'mm': '2', 'ss': '3',
                     'tz': TimezoneTuple(False, False, '4', '5', 'tz name')})

        datetimetest = (DatetimeTuple(DateTuple('1', '2', '3',
                                                '4', '5', '6'),
                                      TimeTuple('7', '8', '9',
                                                TimezoneTuple(True, False,
                                                              '10', '11',
                                                              'tz name'))),
                        (DateTuple('1', '2', '3', '4', '5', '6'),
                         TimeTuple('7', '8', '9',
                                   TimezoneTuple(True, False, '10', '11',
                                                 'tz name'))))

        durationtest = (DurationTuple('1', '2', '3', '4', '5', '6', '7'),
                        {'PnY': '1', 'PnM': '2', 'PnW': '3', 'PnD': '4',
                         'TnH': '5', 'TnM': '6', 'TnS': '7'})

        intervaltests = ((IntervalTuple(DateTuple('1', '2', '3', '4',
                                                  '5', '6'),
                                        DateTuple('7', '8', '9', '10',
                                                  '11', '12'),
                                        None),
                          {'start': DateTuple('1', '2', '3', '4', '5', '6'),
                           'end': DateTuple('7', '8', '9', '10', '11', '12'),
                           'duration': None}),
                         (IntervalTuple(DateTuple('1', '2', '3', '4', '5', '6'),
                                        None,
                                        DurationTuple('7', '8', '9', '10',
                                                      '11', '12', '13')),
                          {'start': DateTuple('1', '2', '3', '4', '5', '6'),
                           'end': None,
                           'duration': DurationTuple('7', '8', '9', '10',
                                                     '11', '12', '13')}),
                         (IntervalTuple(None,
                                        TimeTuple('1', '2', '3',
                                                  TimezoneTuple(True, False,
                                                                '4', '5',
                                                                'tz name')),
                                        DurationTuple('6', '7', '8', '9', '10', '11', '12')),
                          {'start': None,
                           'end': TimeTuple('1', '2', '3',
                                   TimezoneTuple(True, False, '4', '5',
                                                 'tz name')),
                           'duration': DurationTuple('6', '7', '8', '9',
                                                     '10', '11', '12')}))

        repeatingintervaltests = ((RepeatingIntervalTuple(True,
                                                          None,
                                                          IntervalTuple(DateTuple('1', '2', '3',
                                                                                  '4', '5', '6'),
                                                                        DateTuple('7', '8', '9',
                                                                                  '10', '11', '12'),
                                                                        None)),
                                   {'R': True,
                                    'Rnn': None,
                                    'interval': IntervalTuple(DateTuple('1', '2', '3',
                                                                        '4', '5', '6'),
                                                              DateTuple('7', '8', '9',
                                                                        '10', '11', '12'),
                                                                        None)}),
                                  (RepeatingIntervalTuple(False,
                                                          '1',
                                                          IntervalTuple(DatetimeTuple(DateTuple('2', '3', '4',
                                                                                                '5', '6', '7'),
                                                                                      TimeTuple('8', '9', '10',
                                                                                                None)),
                                                                        DatetimeTuple(DateTuple('11', '12', '13',
                                                                                                '14', '15', '16'),
                                                                                      TimeTuple('17', '18', '19',
                                                                                                None)),
                                                                        None)),
                                   {'R':False,
                                    'Rnn': '1',
                                    'interval': IntervalTuple(DatetimeTuple(DateTuple('2', '3', '4',
                                                                                      '5', '6', '7'),
                                                                            TimeTuple('8', '9', '10', None)),
                                                              DatetimeTuple(DateTuple('11', '12', '13',
                                                                                      '14', '15', '16'),
                                                                            TimeTuple('17', '18', '19', None)),
                                                              None)}))

        timezonetest = (TimezoneTuple(False, False, '1', '2', '+01:02'),
                        {'negative': False, 'Z': False,
                         'hh': '1', 'mm': '2', 'name': '+01:02'})

        with mock.patch.object(aniso8601.builders.BaseTimeBuilder,
                               'build_date') as mock_build:
            mock_build.return_value = datetest[0]

            result = BaseTimeBuilder._build_object(datetest[0])

            self.assertEqual(result, datetest[0])
            mock_build.assert_called_once_with(**datetest[1])

        with mock.patch.object(aniso8601.builders.BaseTimeBuilder,
                               'build_time') as mock_build:
            mock_build.return_value = timetest[0]

            result = BaseTimeBuilder._build_object(timetest[0])

            self.assertEqual(result, timetest[0])
            mock_build.assert_called_once_with(**timetest[1])

        with mock.patch.object(aniso8601.builders.BaseTimeBuilder,
                               'build_datetime') as mock_build:
            mock_build.return_value = datetimetest[0]

            result = BaseTimeBuilder._build_object(datetimetest[0])

            self.assertEqual(result, datetimetest[0])
            mock_build.assert_called_once_with(*datetimetest[1])

        with mock.patch.object(aniso8601.builders.BaseTimeBuilder,
                               'build_duration') as mock_build:
            mock_build.return_value = durationtest[0]

            result = BaseTimeBuilder._build_object(durationtest[0])

            self.assertEqual(result, durationtest[0])
            mock_build.assert_called_once_with(**durationtest[1])

        for intervaltest in intervaltests:
            with mock.patch.object(aniso8601.builders.BaseTimeBuilder,
                                   'build_interval') as mock_build:
                mock_build.return_value = intervaltest[0]

                result = BaseTimeBuilder._build_object(intervaltest[0])

                self.assertEqual(result, intervaltest[0])
                mock_build.assert_called_once_with(**intervaltest[1])

        for repeatingintervaltest in repeatingintervaltests:
            with mock.patch.object(aniso8601.builders.BaseTimeBuilder,
                                   'build_repeating_interval') as mock_build:
                mock_build.return_value = repeatingintervaltest[0]

                result = BaseTimeBuilder._build_object(repeatingintervaltest[0])

                self.assertEqual(result, repeatingintervaltest[0])
                mock_build.assert_called_once_with(**repeatingintervaltest[1])

        with mock.patch.object(aniso8601.builders.BaseTimeBuilder,
                               'build_timezone') as mock_build:
            mock_build.return_value = timezonetest[0]

            result = BaseTimeBuilder._build_object(timezonetest[0])

            self.assertEqual(result, timezonetest[0])
            mock_build.assert_called_once_with(**timezonetest[1])

    def test_is_interval_end_concise(self):
        self.assertTrue(BaseTimeBuilder._is_interval_end_concise(TimeTuple('1', '2', '3',
                                                                           None)))
        self.assertTrue(BaseTimeBuilder._is_interval_end_concise(DateTuple(None, '2', '3',
                                                                           '4', '5', '6')))
        self.assertTrue(BaseTimeBuilder._is_interval_end_concise(DatetimeTuple(DateTuple(None, '2', '3',
                                                                                         '4', '5', '6'),
                                                                               TimeTuple('7', '8', '9',
                                                                                         None))))

        self.assertFalse(BaseTimeBuilder._is_interval_end_concise(DateTuple('1', '2', '3',
                                                                            '4', '5', '6')))
        self.assertFalse(BaseTimeBuilder._is_interval_end_concise(DatetimeTuple(DateTuple('1', '2', '3',
                                                                                          '4', '5', '6'),
                                                                                TimeTuple('7', '8', '9',
                                                                                          None))))

    def test_combine_concise_interval_tuples(self):
        testtuples= ((DateTuple('2020', '01', '01',
                                None, None, None),
                      DateTuple(None, None, '02',
                                None, None, None),
                      DateTuple('2020', '01', '02',
                                None, None, None)),
                     (DateTuple('2008', '02', '15',
                                None, None, None),
                      DateTuple(None, '03', '14',
                                None, None, None),
                      DateTuple('2008', '03', '14',
                                None, None, None)),
                     (DatetimeTuple(DateTuple('2007', '12', '14',
                                              None, None, None),
                                     TimeTuple('13', '30', None,
                                               None)),
                      TimeTuple('15', '30', None,
                                None),
                      DatetimeTuple(DateTuple('2007', '12', '14',
                                              None, None, None),
                                    TimeTuple('15', '30', None,
                                              None))),
                     (DatetimeTuple(DateTuple('2007', '11', '13',
                                              None, None, None),
                                    TimeTuple('09', '00', None,
                                              None)),
                      DatetimeTuple(DateTuple(None, None, '15',
                                              None, None, None),
                                    TimeTuple('17', '00', None,
                                              None)),
                      DatetimeTuple(DateTuple('2007', '11', '15',
                                              None, None, None),
                                    TimeTuple('17', '00', None,
                                              None))),
                     (DatetimeTuple(DateTuple('2007', '11', '13',
                                              None, None, None),
                                    TimeTuple('00', '00', None,
                                              None)),
                      DatetimeTuple(DateTuple(None, None, '16',
                                              None, None, None),
                                    TimeTuple('00', '00', None,
                                              None)),
                      DatetimeTuple(DateTuple('2007', '11', '16',
                                              None, None, None),
                                    TimeTuple('00', '00', None,
                                              None))),
                     (DatetimeTuple(DateTuple('2007', '11', '13',
                                              None, None, None),
                                    TimeTuple('09', '00', None,
                                              TimezoneTuple(False,
                                                            True,
                                                            None,
                                                            None,
                                                            'Z'))),
                      DatetimeTuple(DateTuple(None, None, '15',
                                              None, None, None),
                                    TimeTuple('17', '00', None,
                                              None)),
                     DatetimeTuple(DateTuple('2007', '11', '15',
                                              None, None, None),
                                    TimeTuple('17', '00', None,
                                              TimezoneTuple(False,
                                                            True,
                                                            None,
                                                            None,
                                                            'Z')))))

        for testtuple in testtuples:
            result = BaseTimeBuilder._combine_concise_interval_tuples(testtuple[0], testtuple[1])
            self.assertEqual(result, testtuple[2])

class TestTupleBuilder(unittest.TestCase):
    def test_build_date(self):
        datetuple = TupleBuilder.build_date()

        self.assertEqual(datetuple, DateTuple(None, None, None,
                                              None, None, None))

        datetuple = TupleBuilder.build_date(YYYY='1', MM='2', DD='3',
                                            Www='4', D='5', DDD='6')

        self.assertEqual(datetuple, DateTuple('1', '2', '3',
                                              '4', '5', '6'))

    def test_build_time(self):
        testtuples = (({}, TimeTuple(None, None, None, None)),
                      ({'hh': '1', 'mm': '2', 'ss': '3', 'tz': None},
                       TimeTuple('1', '2', '3', None)),
                      ({'hh': '1', 'mm': '2', 'ss': '3', 'tz': TimezoneTuple(False, False,
                                                                             '4', '5',
                                                                             'tz name')},
                       TimeTuple('1', '2', '3', TimezoneTuple(False, False, '4', '5',
                                                              'tz name'))))

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_time(**testtuple[0]),
                             testtuple[1])

    def test_build_datetime(self):
        testtuples = (({'date': DateTuple('1', '2', '3', '4', '5', '6'),
                        'time': TimeTuple('7', '8', '9', None)},
                       DatetimeTuple(DateTuple('1', '2', '3', '4', '5', '6'),
                                     TimeTuple('7', '8', '9', None))),
                      ({'date': DateTuple('1', '2', '3', '4', '5', '6'),
                        'time': TimeTuple('7', '8', '9',
                                          TimezoneTuple(True, False,
                                                        '10', '11',
                                                        'tz name'))},
                       DatetimeTuple(DateTuple('1', '2', '3', '4', '5', '6'),
                                     TimeTuple('7', '8', '9',
                                               TimezoneTuple(True, False,
                                                             '10', '11',
                                                             'tz name')))))

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_datetime(**testtuple[0]),
                             testtuple[1])

    def test_build_duration(self):
        testtuples = (({}, DurationTuple(None, None, None, None,
                                         None, None, None)),
                      ({'PnY': '1', 'PnM': '2', 'PnW': '3', 'PnD': '4',
                        'TnH': '5', 'TnM': '6', 'TnS': '7'},
                       DurationTuple('1', '2', '3', '4',
                                     '5', '6', '7')))

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_duration(**testtuple[0]),
                             testtuple[1])

    def test_build_interval(self):
        testtuples = (({}, IntervalTuple(None, None, None)),
                      ({'start': DateTuple('1', '2', '3', '4', '5', '6'),
                        'end': DateTuple('7', '8', '9', '10', '11', '12')},
                       IntervalTuple(DateTuple('1', '2', '3', '4', '5', '6'),
                                     DateTuple('7', '8', '9', '10', '11', '12'),
                                               None)),
                      ({'start': TimeTuple('1', '2', '3',
                                           TimezoneTuple(True, False,
                                                         '7', '8',
                                                         'tz name')),
                        'end': TimeTuple('4', '5', '6',
                                         TimezoneTuple(False, False,
                                                       '9', '10',
                                                       'tz name'))},
                       IntervalTuple(TimeTuple('1', '2', '3',
                                               TimezoneTuple(True, False,
                                                             '7', '8',
                                                             'tz name')),
                                     TimeTuple('4', '5', '6',
                                               TimezoneTuple(False, False,
                                                             '9', '10',
                                                             'tz name')),
                                     None)),
                      ({'start': DatetimeTuple(DateTuple('1', '2', '3', '4', '5', '6'),
                                               TimeTuple('7', '8', '9',
                                               TimezoneTuple(True, False,
                                                             '10', '11',
                                                             'tz name'))),
                        'end': DatetimeTuple(DateTuple('12', '13', '14',
                                                       '15', '16', '17'),
                                             TimeTuple('18', '19', '20',
                                             TimezoneTuple(False, False,
                                                           '21', '22',
                                                           'tz name')))},
                       IntervalTuple(DatetimeTuple(DateTuple('1', '2', '3',
                                                             '4', '5', '6'),
                                                   TimeTuple('7', '8', '9',
                                                   TimezoneTuple(True, False,
                                                                 '10', '11',
                                                                 'tz name'))),
                                     DatetimeTuple(DateTuple('12', '13', '14',
                                                             '15', '16', '17'),
                                                   TimeTuple('18', '19', '20',
                                                             TimezoneTuple(False,
                                                                           False,
                                                                           '21',
                                                                           '22',
                                                                           'tz name'))),
                        None)),
                      ({'start': DateTuple('1', '2', '3', '4', '5', '6'),
                        'end': None,
                        'duration': DurationTuple('7', '8', '9', '10',
                                                  '11', '12', '13')},
                       IntervalTuple(DateTuple('1', '2', '3', '4', '5', '6'),
                                     None,
                                     DurationTuple('7', '8', '9', '10',
                                                   '11', '12', '13'))),
                      ({'start': None,
                        'end': TimeTuple('1', '2', '3',
                                         TimezoneTuple(True, False, '4', '5',
                                                       'tz name')),
                        'duration': DurationTuple('6', '7', '8', '9',
                                                  '10', '11', '12')},
                       IntervalTuple(None,
                                     TimeTuple('1', '2', '3',
                                               TimezoneTuple(True, False,
                                                             '4', '5',
                                                             'tz name')),
                                     DurationTuple('6', '7', '8', '9', '10',
                                                   '11', '12'))))

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_interval(**testtuple[0]),
                             testtuple[1])

    def test_build_repeating_interval(self):
        testtuples = (({}, RepeatingIntervalTuple(None, None, None)),
                      ({'R': True,
                        'interval': IntervalTuple(DateTuple('1', '2', '3',
                                                            '4', '5', '6'),
                                                  DateTuple('7', '8', '9',
                                                            '10', '11', '12'),
                                                  None)},
                       RepeatingIntervalTuple(True, None, IntervalTuple(DateTuple('1', '2', '3',
                                                                                  '4', '5', '6'),
                                                                        DateTuple('7', '8', '9',
                                                                                  '10', '11', '12'),
                                                                        None))),
                      ({'R':False, 'Rnn': '1',
                        'interval': IntervalTuple(DatetimeTuple(DateTuple('2', '3', '4',
                                                                          '5', '6', '7'),
                                                                TimeTuple('8', '9', '10',
                                                                          None)),
                                                  DatetimeTuple(DateTuple('11', '12', '13',
                                                                          '14', '15', '16'),
                                                                TimeTuple('17', '18', '19',
                                                                          None)),
                                                  None)},
                       RepeatingIntervalTuple(False, '1',
                                              IntervalTuple(DatetimeTuple(DateTuple('2', '3', '4',
                                                                                    '5', '6', '7'),
                                                                          TimeTuple('8', '9', '10',
                                                                                    None)),
                                                            DatetimeTuple(DateTuple('11', '12', '13',
                                                                                    '14', '15', '16'),
                                                                          TimeTuple('17', '18', '19',
                                                                                    None)),
                                                            None))))

        for testtuple in testtuples:
            result = TupleBuilder.build_repeating_interval(**testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_timezone(self):
        testtuples = (({}, TimezoneTuple(None, None, None, None, '')),
                      ({'negative': False, 'Z': True, 'name': 'UTC'},
                       TimezoneTuple(False, True, None, None, 'UTC')),
                      ({'negative': False, 'Z': False, 'hh': '1', 'mm': '2',
                        'name': '+01:02'},
                       TimezoneTuple(False, False, '1', '2', '+01:02')),
                      ({'negative': True, 'Z': False, 'hh': '1', 'mm': '2',
                        'name': '-01:02'},
                       TimezoneTuple(True, False, '1', '2', '-01:02')))

        for testtuple in testtuples:
            result = TupleBuilder.build_timezone(**testtuple[0])
            self.assertEqual(result, testtuple[1])



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.compat import is_string
from aniso8601.exceptions import ISOFormatError

def parse_timezone(tzstr, builder=PythonTimeBuilder):
    #tzstr can be Z, ±hh:mm, ±hhmm, ±hh
    if is_string(tzstr) is False:
        raise ValueError('Time zone must be string.')

    if len(tzstr) == 1 and tzstr[0] == 'Z':
        return builder.build_timezone(negative=False, Z=True, name=tzstr)
    elif len(tzstr) == 6:
        #±hh:mm
        hourstr = tzstr[1:3]
        minutestr = tzstr[4:6]

        if tzstr[0] == '-' and hourstr == '00' and minutestr == '00':
            raise ISOFormatError('Negative ISO 8601 time offset must not '
                                 'be 0.')
    elif len(tzstr) == 5:
        #±hhmm
        hourstr = tzstr[1:3]
        minutestr = tzstr[3:5]

        if tzstr[0] == '-' and hourstr == '00' and minutestr == '00':
            raise ISOFormatError('Negative ISO 8601 time offset must not '
                                 'be 0.')
    elif len(tzstr) == 3:
        #±hh
        hourstr = tzstr[1:3]
        minutestr = None

        if tzstr[0] == '-' and hourstr == '00':
            raise ISOFormatError('Negative ISO 8601 time offset must not '
                                 'be 0.')
    else:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 time offset.'
                             .format(tzstr))

    for componentstr in [hourstr, minutestr]:
        if componentstr is not None:
            if componentstr.isdigit() is False:
                raise ISOFormatError('"{0}" is not a valid ISO 8601 time offset.'
                                     .format(tzstr))

    if tzstr[0] == '+':
        return builder.build_timezone(negative=False, hh=hourstr,
                                      mm=minutestr, name=tzstr)

    if tzstr[0] == '-':
        return builder.build_timezone(negative=True, hh=hourstr,
                                      mm=minutestr, name=tzstr)

    raise ISOFormatError('"{0}" is not a valid ISO 8601 time offset.'
                         .format(tzstr))



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import datetime
import unittest

from aniso8601 import compat
from aniso8601.exceptions import (DayOutOfBoundsError, HoursOutOfBoundsError,
                                  ISOFormatError, LeapSecondError,
                                  MidnightBoundsError, MinutesOutOfBoundsError,
                                  MonthOutOfBoundsError,
                                  SecondsOutOfBoundsError,
                                  WeekOutOfBoundsError, YearOutOfBoundsError)
from aniso8601.builders import (DateTuple, DatetimeTuple, DurationTuple,
                                IntervalTuple, Limit, TimeTuple,
                                TimezoneTuple)
from aniso8601.builders.python import (_cast_to_fractional_component,
                                       FractionalComponent, PythonTimeBuilder,
                                       fractional_range_check, year_range_check)
from aniso8601.utcoffset import UTCOffset

class TestPythonTimeBuilder_UtiltyFunctions(unittest.TestCase):
    def test_year_range_check(self):
        yearlimit = Limit('Invalid year string.',
                          0000, 9999, YearOutOfBoundsError,
                          'Year must be between 1..9999.',
                          None)

        self.assertEqual(year_range_check('1', yearlimit), 1000)

    def test_fractional_range_check(self):
        limit = Limit('Invalid string.',
                      -1, 1, ValueError,
                      'Value must be between -1..1.',
                      None)

        self.assertEqual(fractional_range_check(10, '1', limit), 1)
        self.assertEqual(fractional_range_check(10, '-1', limit), -1)
        self.assertEqual(fractional_range_check(10, '0.1', limit), FractionalComponent(0, 1))
        self.assertEqual(fractional_range_check(10, '-0.1', limit), FractionalComponent(-0, 1))

        with self.assertRaises(ValueError):
            fractional_range_check(10, '1.1', limit)

        with self.assertRaises(ValueError):
            fractional_range_check(10, '-1.1', limit)

    def test_cast_to_fractional_component(self):
        self.assertEqual(_cast_to_fractional_component(10, '1.1'), FractionalComponent(1, 1))
        self.assertEqual(_cast_to_fractional_component(10, '-1.1'), FractionalComponent(-1, 1))

        self.assertEqual(_cast_to_fractional_component(100, '1.1'), FractionalComponent(1, 10))
        self.assertEqual(_cast_to_fractional_component(100, '-1.1'), FractionalComponent(-1, 10))

class TestPythonTimeBuilder(unittest.TestCase):
    def test_build_date(self):
        testtuples = (({'YYYY': '2013', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(2013, 1, 1)),
                      ({'YYYY': '0001', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(1, 1, 1)),
                      ({'YYYY': '1900', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(1900, 1, 1)),
                      ({'YYYY': '1981', 'MM': '04', 'DD': '05', 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(1981, 4, 5)),
                      ({'YYYY': '1981', 'MM': '04', 'DD': None, 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(1981, 4, 1)),
                      ({'YYYY': '1981', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': '095'},
                       datetime.date(1981, 4, 5)),
                      ({'YYYY': '1981', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': '365'},
                       datetime.date(1981, 12, 31)),
                      ({'YYYY': '1980', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': '366'},
                       datetime.date(1980, 12, 31)),
                      #Make sure we shift in zeros
                      ({'YYYY': '1', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(1000, 1, 1)),
                      ({'YYYY': '12', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(1200, 1, 1)),
                      ({'YYYY': '123', 'MM': None, 'DD': None, 'Www': None,
                        'D': None, 'DDD': None},
                       datetime.date(1230, 1, 1)))

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_date(**testtuple[0])
            self.assertEqual(result, testtuple[1])

        #Test weekday
        testtuples = (({'YYYY': '2004', 'MM': None, 'DD': None, 'Www': '53',
                        'D': None, 'DDD': None},
                       datetime.date(2004, 12, 27), 0),
                      ({'YYYY': '2009', 'MM': None, 'DD': None, 'Www': '01',
                        'D': None, 'DDD': None},
                       datetime.date(2008, 12, 29), 0),
                      ({'YYYY': '2010', 'MM': None, 'DD': None, 'Www': '01',
                        'D': None, 'DDD': None},
                       datetime.date(2010, 1, 4), 0),
                      ({'YYYY': '2009', 'MM': None, 'DD': None, 'Www': '53',
                        'D': None, 'DDD': None},
                       datetime.date(2009, 12, 28), 0),
                      ({'YYYY': '2009', 'MM': None, 'DD': None, 'Www': '01',
                        'D': '1', 'DDD': None},
                       datetime.date(2008, 12, 29), 0),
                      ({'YYYY': '2009', 'MM': None, 'DD': None, 'Www': '53',
                        'D': '7', 'DDD': None},
                       datetime.date(2010, 1, 3), 6),
                      ({'YYYY': '2010', 'MM': None, 'DD': None, 'Www': '01',
                        'D': '1', 'DDD': None},
                       datetime.date(2010, 1, 4), 0),
                      ({'YYYY': '2004', 'MM': None, 'DD': None, 'Www': '53',
                        'D': '6', 'DDD': None},
                       datetime.date(2005, 1, 1), 5))

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_date(**testtuple[0])
            self.assertEqual(result, testtuple[1])
            self.assertEqual(result.weekday(), testtuple[2])

    def test_build_time(self):
        testtuples = (({}, datetime.time()),
                      ({'hh': '12.5'},
                       datetime.time(hour=12, minute=30)),
                      ({'hh': '23.99999999997'},
                       datetime.time(hour=23, minute=59, second=59,
                                     microsecond=999999)),
                      ({'hh': '1', 'mm': '23'},
                       datetime.time(hour=1, minute=23)),
                      ({'hh': '1', 'mm': '23.4567'},
                       datetime.time(hour=1, minute=23, second=27,
                                     microsecond=402000)),
                      ({'hh': '14', 'mm': '43.999999997'},
                       datetime.time(hour=14, minute=43, second=59,
                                     microsecond=999999)),
                      ({'hh': '1', 'mm': '23', 'ss': '45'},
                       datetime.time(hour=1, minute=23, second=45)),
                      ({'hh': '23', 'mm': '21', 'ss': '28.512400'},
                       datetime.time(hour=23, minute=21, second=28,
                                     microsecond=512400)),
                      ({'hh': '01', 'mm': '03', 'ss': '11.858714'},
                       datetime.time(hour=1, minute=3, second=11,
                                     microsecond=858714)),
                      ({'hh': '14', 'mm': '43', 'ss': '59.9999997'},
                       datetime.time(hour=14, minute=43, second=59,
                                     microsecond=999999)),
                      ({'hh': '24'}, datetime.time(hour=0)),
                      ({'hh': '24', 'mm': '00'}, datetime.time(hour=0)),
                      ({'hh': '24', 'mm': '00', 'ss': '00'},
                       datetime.time(hour=0)),
                      ({'tz': TimezoneTuple(False, None, '00', '00', 'UTC')},
                       datetime.time(tzinfo=UTCOffset(name='UTC', minutes=0))),
                      ({'hh': '23', 'mm': '21', 'ss': '28.512400',
                        'tz': TimezoneTuple(False, None, '00', '00', '+00:00')},
                       datetime.time(hour=23, minute=21, second=28,
                                     microsecond=512400,
                                     tzinfo=UTCOffset(name='+00:00',
                                                      minutes=0))),
                      ({'hh': '1', 'mm': '23',
                        'tz': TimezoneTuple(False, None, '01', '00', '+1')},
                       datetime.time(hour=1, minute=23,
                                     tzinfo=UTCOffset(name='+1',
                                                      minutes=60))),
                      ({'hh': '1', 'mm': '23.4567',
                        'tz': TimezoneTuple(True, None, '01', '00', '-1')},
                       datetime.time(hour=1, minute=23, second=27,
                                     microsecond=402000,
                                     tzinfo=UTCOffset(name='-1',
                                                      minutes=-60))),
                      ({'hh': '23', 'mm': '21', 'ss': '28.512400',
                        'tz': TimezoneTuple(False, None, '01', '30', '+1:30')},
                       datetime.time(hour=23, minute=21, second=28,
                                     microsecond=512400,
                                     tzinfo=UTCOffset(name='+1:30',
                                                      minutes=90))),
                      ({'hh': '23', 'mm': '21', 'ss': '28.512400',
                        'tz': TimezoneTuple(False, None, '11', '15', '+11:15')},
                       datetime.time(hour=23, minute=21, second=28,
                                     microsecond=512400,
                                     tzinfo=UTCOffset(name='+11:15',
                                                      minutes=675))),
                      ({'hh': '23', 'mm': '21', 'ss': '28.512400',
                        'tz': TimezoneTuple(False, None, '12', '34', '+12:34')},
                       datetime.time(hour=23, minute=21, second=28,
                                     microsecond=512400,
                                     tzinfo=UTCOffset(name='+12:34',
                                                      minutes=754))),
                      ({'hh': '23', 'mm': '21', 'ss': '28.512400',
                        'tz': TimezoneTuple(False, None, '00', '00', 'UTC')},
                       datetime.time(hour=23, minute=21, second=28,
                                     microsecond=512400,
                                     tzinfo=UTCOffset(name='UTC',
                                                      minutes=0))),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      #https://bitbucket.org/nielsenb/aniso8601/issues/21/sub-microsecond-precision-is-lost-when
                      ({'hh': '14.9999999999999999'},
                       datetime.time(hour=14, minute=59, second=59,
                                     microsecond=999999)),
                      ({'mm': '0.00000000999'},
                       datetime.time()),
                      ({'mm': '0.0000000999'},
                       datetime.time(microsecond=5)),
                      ({'ss': '0.0000001'},
                       datetime.time()),
                      ({'ss': '2.0000048'},
                       datetime.time(second=2,
                                     microsecond=4)))

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_time(**testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_datetime(self):
        testtuples = (((DateTuple('2019', '06', '05', None, None, None),
                        TimeTuple('01', '03', '11.858714', None)),
                       datetime.datetime(2019, 6, 5, hour=1, minute=3,
                                         second=11, microsecond=858714)),
                      ((DateTuple('1234', '02', '03', None, None, None),
                        TimeTuple('23', '21', '28.512400', None)),
                       datetime.datetime(1234, 2, 3, hour=23, minute=21,
                                         second=28, microsecond=512400)),
                      ((DateTuple('1981', '04', '05', None, None, None),
                        TimeTuple('23', '21', '28.512400',
                                  TimezoneTuple(False, None, '11', '15', '+11:15'))),
                       datetime.datetime(1981, 4, 5, hour=23, minute=21,
                                         second=28, microsecond=512400,
                                         tzinfo=UTCOffset(name='+11:15',
                                                          minutes=675))))

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_datetime(*testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_duration(self):
        testtuples = (({'PnY': '1', 'PnM': '2', 'PnD': '3',
                        'TnH': '4', 'TnM': '54', 'TnS': '6'},
                       datetime.timedelta(days=428, hours=4,
                                          minutes=54, seconds=6)),
                      ({'PnY': '1', 'PnM': '2', 'PnD': '3',
                        'TnH': '4', 'TnM': '54', 'TnS': '6.5'},
                       datetime.timedelta(days=428, hours=4,
                                          minutes=54, seconds=6.5)),
                      ({'PnY': '1', 'PnM': '2', 'PnD': '3'},
                       datetime.timedelta(days=428)),
                      ({'PnY': '1', 'PnM': '2', 'PnD': '3.5'},
                       datetime.timedelta(days=428.5)),
                      ({'TnH': '4', 'TnM': '54', 'TnS': '6.5'},
                       datetime.timedelta(hours=4, minutes=54, seconds=6.5)),
                      ({'TnH': '1', 'TnM': '3', 'TnS': '11.858714'},
                       datetime.timedelta(hours=1, minutes=3,
                                          seconds=11, microseconds=858714)),
                      ({'TnH': '4', 'TnM': '54', 'TnS': '28.512400'},
                       datetime.timedelta(hours=4, minutes=54,
                                          seconds=28, microseconds=512400)),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      #https://bitbucket.org/nielsenb/aniso8601/issues/21/sub-microsecond-precision-is-lost-when
                      ({'PnY': '1999.9999999999999999'},
                       datetime.timedelta(days=729999, seconds=86399,
                                          microseconds=999999)),
                      ({'PnM': '1.9999999999999999'},
                       datetime.timedelta(days=59, hours=23,
                                          minutes=59, seconds=59,
                                          microseconds=999999)),
                      ({'PnW': '1.9999999999999999'},
                       datetime.timedelta(days=13, hours=23,
                                          minutes=59, seconds=59,
                                          microseconds=999999)),
                      ({'PnD': '1.9999999999999999'},
                       datetime.timedelta(days=1, hours=23,
                                          minutes=59, seconds=59,
                                          microseconds=999999)),
                      ({'TnH': '14.9999999999999999'},
                       datetime.timedelta(hours=14, minutes=59,
                                          seconds=59, microseconds=999999)),
                      ({'TnM': '0.00000000999'}, datetime.timedelta(0)),
                      ({'TnM': '0.0000000999'},
                       datetime.timedelta(microseconds=5)),
                      ({'TnS': '0.0000001'}, datetime.timedelta(0)),
                      ({'TnS': '2.0000048'},
                       datetime.timedelta(seconds=2, microseconds=4)),
                      ({'PnY': '1'}, datetime.timedelta(days=365)),
                      ({'PnY': '1.5'}, datetime.timedelta(days=547.5)),
                      ({'PnM': '1'}, datetime.timedelta(days=30)),
                      ({'PnM': '1.5'}, datetime.timedelta(days=45)),
                      ({'PnW': '1'}, datetime.timedelta(days=7)),
                      ({'PnW': '1.5'}, datetime.timedelta(days=10.5)),
                      ({'PnD': '1'}, datetime.timedelta(days=1)),
                      ({'PnD': '1.5'}, datetime.timedelta(days=1.5)),
                      ({'PnY': '0003', 'PnM': '06', 'PnD': '04',
                        'TnH': '12', 'TnM': '30', 'TnS': '05'},
                       datetime.timedelta(days=1279, hours=12,
                                          minutes=30, seconds=5)),
                      ({'PnY': '0003', 'PnM': '06', 'PnD': '04',
                        'TnH': '12', 'TnM': '30', 'TnS': '05.5'},
                       datetime.timedelta(days=1279, hours=12,
                                          minutes=30, seconds=5.5)),
                      #Test timedelta limit
                      ({'PnD': '999999999', 'TnH': '23', 'TnM': '59',
                        'TnS': '59.999999'},
                       datetime.timedelta.max),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ({'PnY': '0001', 'PnM': '02', 'PnD': '03',
                        'TnH': '14', 'TnM': '43', 'TnS': '59.9999997'},
                       datetime.timedelta(days=428, hours=14,
                                          minutes=43, seconds=59,
                                          microseconds=999999)),
                      #Verify overflows
                      ({'TnH': '36'}, datetime.timedelta(days=1, hours=12)))

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_duration(**testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_interval(self):
        testtuples = (({'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None)),
                        'duration': DurationTuple(None, '1', None, None,
                                                  None, None, None)},
                       datetime.datetime(year=1981, month=4, day=5,
                                         hour=1, minute=1),
                       datetime.datetime(year=1981, month=3, day=6,
                                         hour=1, minute=1)),
                      ({'end': DateTuple('1981', '04', '05', None, None, None),
                        'duration': DurationTuple(None, '1', None, None,
                                                  None, None, None)},
                       datetime.date(year=1981, month=4, day=5),
                       datetime.date(year=1981, month=3, day=6)),
                      ({'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple('1.5', None, None, None,
                                                  None, None, None)},
                       datetime.date(year=2018, month=3, day=6),
                       datetime.datetime(year=2016, month=9, day=4,
                                         hour=12)),
                      ({'end': DateTuple('2014', '11', '12', None, None, None),
                        'duration': DurationTuple(None, None, None, None,
                                                  '1', None, None)},
                       datetime.date(year=2014, month=11, day=12),
                       datetime.datetime(year=2014, month=11, day=11,
                                         hour=23)),
                      ({'end': DateTuple('2014', '11', '12', None, None, None),
                        'duration': DurationTuple(None, None, None, None,
                                                  '4', '54', '6.5')},
                       datetime.date(year=2014, month=11, day=12),
                       datetime.datetime(year=2014, month=11, day=11,
                                         hour=19, minute=5, second=53,
                                         microsecond=500000)),
                      ({'end': DatetimeTuple(DateTuple('2050', '03', '01',
                                                       None, None, None),
                                             TimeTuple('13', '00', '00',
                                                       TimezoneTuple(False,
                                                                     True,
                                                                     None,
                                                                     None,
                                                                     'Z'))),
                        'duration': DurationTuple(None, None, None,
                                                  None, '10', None, None)},
                       datetime.datetime(year=2050, month=3, day=1,
                                         hour=13,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0)),
                       datetime.datetime(year=2050, month=3, day=1,
                                         hour=3,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0))),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      #https://bitbucket.org/nielsenb/aniso8601/issues/21/sub-microsecond-precision-is-lost-when
                      ({'end': DateTuple('2000', '01', '01',
                                         None, None, None),
                        'duration': DurationTuple('1999.9999999999999999', None, None,
                                                  None, None, None,
                                                  None)},
                       datetime.date(year=2000, month=1, day=1),
                       datetime.datetime(year=1, month=4, day=30,
                                         hour=0, minute=0, second=0,
                                         microsecond=1)),
                      ({'end': DateTuple('1989', '03', '01',
                                         None, None, None),
                        'duration': DurationTuple(None, '1.9999999999999999', None,
                                                  None, None, None,
                                                  None)},
                       datetime.date(year=1989, month=3, day=1),
                       datetime.datetime(year=1988, month=12, day=31,
                                         hour=0, minute=0, second=0,
                                         microsecond=1)),
                      ({'end': DateTuple('1989', '03', '01',
                                         None, None, None),
                        'duration': DurationTuple(None, None, '1.9999999999999999',
                                                  None, None, None,
                                                  None)},
                       datetime.date(year=1989, month=3, day=1),
                       datetime.datetime(year=1989, month=2, day=15,
                                         hour=0, minute=0, second=0,
                                         microsecond=1)),
                      ({'end': DateTuple('1989', '03', '01',
                                         None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  '1.9999999999999999', None, None,
                                                  None)},
                       datetime.date(year=1989, month=3, day=1),
                       datetime.datetime(year=1989, month=2, day=27,
                                         hour=0, minute=0, second=0,
                                         microsecond=1)),
                      ({'end': DateTuple('2001', '01', '01',
                                         None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, '14.9999999999999999', None,
                                                  None)},
                       datetime.date(year=2001, month=1, day=1),
                       datetime.datetime(year=2000, month=12, day=31,
                                         hour=9, minute=0, second=0,
                                         microsecond=1)),
                      ({'end': DateTuple('2001', '01', '01',
                                         None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, '0.00000000999',
                                                  None)},
                       datetime.date(year=2001, month=1, day=1),
                       datetime.datetime(year=2001, month=1, day=1)),
                      ({'end': DateTuple('2001', '01', '01',
                                         None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, '0.0000000999',
                                                  None)},
                       datetime.date(year=2001, month=1, day=1),
                       datetime.datetime(year=2000, month=12, day=31,
                                         hour=23, minute=59, second=59,
                                         microsecond=999995)),
                      ({'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '0.0000001')},
                       datetime.date(year=2018, month=3, day=6),
                       datetime.datetime(year=2018, month=3, day=6)),
                      ({'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '2.0000048')},
                       datetime.date(year=2018, month=3, day=6),
                       datetime.datetime(year=2018, month=3, day=5,
                                         hour=23, minute=59, second=57,
                                         microsecond=999996)),
                      ({'start': DatetimeTuple(DateTuple('1981', '04', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'duration': DurationTuple(None, '1', None,
                                                  '1', None, '1', None)},
                       datetime.datetime(year=1981, month=4, day=5,
                                         hour=1, minute=1),
                       datetime.datetime(year=1981, month=5, day=6,
                                         hour=1, minute=2)),
                      ({'start': DateTuple('1981', '04', '05',
                                           None, None, None),
                        'duration': DurationTuple(None, '1', None,
                                                  '1', None, None, None)},
                       datetime.date(year=1981, month=4, day=5),
                       datetime.date(year=1981, month=5, day=6)),
                      ({'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, '2.5', None,
                                                  None, None, None, None)},
                       datetime.date(year=2018, month=3, day=6),
                       datetime.date(year=2018, month=5, day=20)),
                      ({'start': DateTuple('2014', '11', '12',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, '1', None, None)},
                       datetime.date(year=2014, month=11, day=12),
                       datetime.datetime(year=2014, month=11, day=12,
                                         hour=1, minute=0)),
                      ({'start': DateTuple('2014', '11', '12',
                                  None, None, None),
                        'duration': DurationTuple(None, None, None,
                                     None, '4', '54', '6.5')},
                       datetime.date(year=2014, month=11, day=12),
                       datetime.datetime(year=2014, month=11, day=12,
                                         hour=4, minute=54, second=6,
                                         microsecond=500000)),
                      ({'start': DatetimeTuple(DateTuple('2050', '03', '01',
                                                         None, None, None),
                                               TimeTuple('13', '00', '00',
                                                         TimezoneTuple(False,
                                                                       True,
                                                                       None,
                                                                       None,
                                                                       'Z'))),
                        'duration': DurationTuple(None, None, None,
                                                  None, '10', None, None)},
                       datetime.datetime(year=2050, month=3, day=1,
                                         hour=13,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0)),
                       datetime.datetime(year=2050, month=3, day=1,
                                         hour=23,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0))),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ({'start': DateTuple('0001', '01', '01',
                                           None, None, None),
                        'duration': DurationTuple('1999.9999999999999999', None, None,
                                                  None, None, None,
                                                  None)},
                       datetime.date(year=1, month=1, day=1),
                       datetime.datetime(year=1999, month=9, day=3,
                                         hour=23, minute=59, second=59,
                                         microsecond=999999)),
                      ({'start': DateTuple('1989', '03', '01',
                                           None, None, None),
                        'duration': DurationTuple(None, '1.9999999999999999', None,
                                                  None, None, None,
                                                  None)},
                       datetime.date(year=1989, month=3, day=1),
                       datetime.datetime(year=1989, month=4, day=29,
                                         hour=23, minute=59, second=59,
                                         microsecond=999999)),
                      ({'start': DateTuple('1989', '03', '01',
                                           None, None, None),
                        'duration': DurationTuple(None, None, '1.9999999999999999',
                                                  None, None, None,
                                                  None)},
                       datetime.date(year=1989, month=3, day=1),
                       datetime.datetime(year=1989, month=3, day=14,
                                         hour=23, minute=59, second=59,
                                         microsecond=999999)),
                      ({'start': DateTuple('1989', '03', '01',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  '1.9999999999999999', None, None,
                                                  None)},
                       datetime.date(year=1989, month=3, day=1),
                       datetime.datetime(year=1989, month=3, day=2,
                                         hour=23, minute=59, second=59,
                                         microsecond=999999)),
                      ({'start': DateTuple('2001', '01', '01',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, '14.9999999999999999', None,
                                                  None)},
                       datetime.date(year=2001, month=1, day=1),
                       datetime.datetime(year=2001, month=1, day=1,
                                         hour=14, minute=59, second=59,
                                         microsecond=999999)),
                      ({'start': DateTuple('2001', '01', '01',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, '0.00000000999',
                                                  None)},
                       datetime.date(year=2001, month=1, day=1),
                       datetime.datetime(year=2001, month=1, day=1)),
                      ({'start': DateTuple('2001', '01', '01',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, '0.0000000999',
                                                  None)},
                       datetime.date(year=2001, month=1, day=1),
                       datetime.datetime(year=2001, month=1, day=1,
                                         hour=0, minute=0, second=0,
                                         microsecond=5)),
                      ({'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '0.0000001')},
                       datetime.date(year=2018, month=3, day=6),
                       datetime.datetime(year=2018, month=3, day=6)),
                      ({'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '2.0000048')},
                       datetime.date(year=2018, month=3, day=6),
                       datetime.datetime(year=2018, month=3, day=6,
                                         hour=0, minute=0, second=2,
                                         microsecond=4)),
                      ({'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))},
                       datetime.datetime(year=1980, month=3, day=5,
                                         hour=1, minute=1),
                       datetime.datetime(year=1981, month=4, day=5,
                                         hour=1, minute=1)),
                      ({'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'end': DateTuple('1981', '04', '05',
                                         None, None, None)},
                       datetime.datetime(year=1980, month=3, day=5,
                                         hour=1, minute=1),
                       datetime.date(year=1981, month=4, day=5)),
                      ({'start': DateTuple('1980', '03', '05',
                                           None, None, None),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))},
                       datetime.date(year=1980, month=3, day=5),
                       datetime.datetime(year=1981, month=4, day=5,
                                         hour=1, minute=1)),
                      ({'start': DateTuple('1980', '03', '05',
                                           None, None, None),
                        'end': DateTuple('1981', '04', '05',
                                         None, None, None)},
                       datetime.date(year=1980, month=3, day=5),
                       datetime.date(year=1981, month=4, day=5)),
                      ({'start': DateTuple('1981', '04', '05',
                                           None, None, None),
                        'end': DateTuple('1980', '03', '05',
                                         None, None, None)},
                       datetime.date(year=1981, month=4, day=5),
                       datetime.date(year=1980, month=3, day=5)),
                      ({'start': DatetimeTuple(DateTuple('2050', '03', '01',
                                                         None, None, None),
                                               TimeTuple('13', '00', '00',
                                                         TimezoneTuple(False,
                                                                       True,
                                                                       None,
                                                                       None,
                                                                       'Z'))),
                        'end': DatetimeTuple(DateTuple('2050', '05', '11',
                                                       None, None, None),
                                             TimeTuple('15', '30', '00',
                                                       TimezoneTuple(False,
                                                                     True,
                                                                     None,
                                                                     None,
                                                                     'Z')))},
                       datetime.datetime(year=2050, month=3, day=1,
                                         hour=13,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0)),
                       datetime.datetime(year=2050, month=5, day=11,
                                         hour=15, minute=30,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0))),
                      #Test concise representation
                      ({'start': DateTuple('2020', '01', '01',
                                           None, None, None),
                        'end': DateTuple(None, None, '02',
                                         None, None, None)},
                        datetime.date(year=2020, month=1, day=1),
                        datetime.date(year=2020, month=1, day=2)),
                      ({'start': DateTuple('2008', '02', '15',
                                           None, None, None),
                        'end': DateTuple(None, '03', '14',
                                         None, None, None)},
                        datetime.date(year=2008, month=2, day=15),
                        datetime.date(year=2008, month=3, day=14)),
                      ({'start': DatetimeTuple(DateTuple('2007', '12', '14',
                                                         None, None, None),
                                               TimeTuple('13', '30', None,
                                                         None)),
                        'end': TimeTuple('15', '30', None,
                                         None)},
                        datetime.datetime(year=2007, month=12, day=14,
                                          hour=13, minute=30),
                        datetime.datetime(year=2007, month=12, day=14,
                                          hour=15, minute=30)),
                      ({'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('09', '00', None,
                                                         None)),
                        'end': DatetimeTuple(DateTuple(None, None, '15',
                                                       None, None, None),
                                             TimeTuple('17', '00', None,
                                                       None))},
                        datetime.datetime(year=2007, month=11, day=13,
                                          hour=9),
                        datetime.datetime(year=2007, month=11, day=15,
                                          hour=17)),
                      ({'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('00', '00', None,
                                                         None)),
                        'end': DatetimeTuple(DateTuple(None, None, '16',
                                                       None, None, None),
                                             TimeTuple('00', '00', None,
                                                       None))},
                        datetime.datetime(year=2007, month=11, day=13),
                        datetime.datetime(year=2007, month=11, day=16)),
                      ({'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('09', '00', None,
                                                         TimezoneTuple(False,
                                                                       True,
                                                                       None,
                                                                       None,
                                                                       'Z'))),
                        'end': DatetimeTuple(DateTuple(None, None, '15',
                                                       None, None, None),
                                             TimeTuple('17', '00', None,
                                                       None))},
                       datetime.datetime(year=2007, month=11, day=13,
                                         hour=9,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0)),
                       datetime.datetime(year=2007, month=11, day=15,
                                         hour=17,
                                         tzinfo=UTCOffset(name='UTC',
                                                          minutes=0))),
                      ({'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('09', '00', None,
                                                         None)),
                        'end': TimeTuple('12', '34.567', None,
                                         None)},
                       datetime.datetime(year=2007, month=11, day=13,
                                         hour=9),
                       datetime.datetime(year=2007, month=11, day=13,
                                         hour=12, minute=34, second=34,
                                         microsecond=20000)),
                      ({'start': DateTuple('2007', '11', '13',
                                           None, None, None),
                        'end': TimeTuple('12', '34', None,
                                         None)},
                        datetime.date(year=2007, month=11, day=13),
                        datetime.datetime(year=2007, month=11, day=13,
                                          hour=12, minute=34)),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ({'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00.0000001',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('14', '43',
                                                       '59.9999997', None))},
                       datetime.datetime(year=1980, month=3, day=5,
                                         hour=1, minute=1),
                       datetime.datetime(year=1981, month=4, day=5,
                                         hour=14, minute=43, second=59,
                                         microsecond=999999)))

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_interval(**testtuple[0])
            self.assertEqual(result[0], testtuple[1])
            self.assertEqual(result[1], testtuple[2])

    def test_build_repeating_interval(self):
        args = {'Rnn': '3', 'interval': IntervalTuple(DateTuple('1981', '04', '05',
                                                                None, None, None),
                                                      None,
                                                      DurationTuple(None, None, None,
                                                                    '1', None, None,
                                                                    None))}
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        self.assertEqual(results[0], datetime.date(year=1981, month=4, day=5))
        self.assertEqual(results[1], datetime.date(year=1981, month=4, day=6))
        self.assertEqual(results[2], datetime.date(year=1981, month=4, day=7))

        args = {'Rnn': '11', 'interval': IntervalTuple(None,
                                                       DatetimeTuple(DateTuple('1980', '03', '05',
                                                                               None, None, None),
                                                                     TimeTuple('01', '01', '00',
                                                                               None)),
                                                                     DurationTuple(None, None, None,
                                                                                   None, '1', '2',
                                                                                   None))}
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        for dateindex in compat.range(0, 11):
            self.assertEqual(results[dateindex],
                             datetime.datetime(year=1980, month=3, day=5,
                                               hour=1, minute=1)
                             - dateindex * datetime.timedelta(hours=1,
                                                              minutes=2))

        args = {'Rnn': '2', 'interval': IntervalTuple(DatetimeTuple(DateTuple('1980', '03', '05',
                                                                              None, None, None),
                                                                    TimeTuple('01', '01', '00',
                                                                              None)),
                                                      DatetimeTuple(DateTuple('1981', '04', '05',
                                                                              None, None, None),
                                                                    TimeTuple('01', '01', '00',
                                                                              None)),
                                                      None)}
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        self.assertEqual(results[0],
                         datetime.datetime(year=1980, month=3, day=5,
                                           hour=1, minute=1))
        self.assertEqual(results[1],
                         datetime.datetime(year=1981, month=4, day=5,
                                           hour=1, minute=1))

        args = {'Rnn': '2', 'interval': IntervalTuple(DatetimeTuple(DateTuple('1980', '03', '05',
                                                                              None, None, None),
                                                                    TimeTuple('01', '01', '00',
                                                                              None)),
                                                      DatetimeTuple(DateTuple('1981', '04', '05',
                                                                              None, None, None),
                                                                    TimeTuple('01', '01', '00',
                                                                              None)),
                                                      None)}
        results = list(PythonTimeBuilder.build_repeating_interval(**args))

        self.assertEqual(results[0],
                         datetime.datetime(year=1980, month=3, day=5,
                                           hour=1, minute=1))
        self.assertEqual(results[1],
                         datetime.datetime(year=1981, month=4, day=5,
                                           hour=1, minute=1))

        args = {'R': True, 'interval': IntervalTuple(None,
                                                     DatetimeTuple(DateTuple('1980', '03', '05',
                                                                             None, None, None),
                                                                   TimeTuple('01', '01', '00',
                                                                             None)),
                                                     DurationTuple(None, None, None,
                                                                   None, '1', '2', None))}
        resultgenerator = PythonTimeBuilder.build_repeating_interval(**args)

        #Test the first 11 generated
        for dateindex in compat.range(0, 11):
            self.assertEqual(next(resultgenerator),
                             datetime.datetime(year=1980, month=3, day=5,
                                               hour=1, minute=1)
                             - dateindex * datetime.timedelta(hours=1,
                                                              minutes=2))

        args = {'R': True, 'interval': IntervalTuple(DateTuple('1981', '04', '05',
                                                                None, None, None),
                                                      None,
                                                      DurationTuple(None, None, None,
                                                                    '1', None, None,
                                                                    None))}
        resultgenerator = PythonTimeBuilder.build_repeating_interval(**args)

        #Test the first 11 generated
        for dateindex in compat.range(0, 11):
            self.assertEqual(next(resultgenerator),
                             (datetime.datetime(year=1981, month=4, day=5,
                                               hour=0, minute=0)
                             + dateindex * datetime.timedelta(days=1)).date())

    def test_build_timezone(self):
        testtuples = (({'Z': True, 'name': 'Z'},
                       datetime.timedelta(hours=0), 'UTC'),
                      ({'negative': False, 'hh': '00', 'mm': '00',
                        'name': '+00:00'},
                       datetime.timedelta(hours=0), '+00:00'),
                      ({'negative': False, 'hh': '01', 'mm': '00',
                        'name': '+01:00'},
                       datetime.timedelta(hours=1), '+01:00'),
                      ({'negative': True, 'hh': '01', 'mm': '00',
                        'name': '-01:00'},
                       -datetime.timedelta(hours=1), '-01:00'),
                      ({'negative': False, 'hh': '00', 'mm': '12',
                        'name': '+00:12'},
                       datetime.timedelta(minutes=12), '+00:12'),
                      ({'negative': False, 'hh': '01', 'mm': '23',
                        'name': '+01:23'},
                       datetime.timedelta(hours=1, minutes=23), '+01:23'),
                      ({'negative': True, 'hh': '01', 'mm': '23',
                        'name': '-01:23'},
                       -datetime.timedelta(hours=1, minutes=23), '-01:23'),
                      ({'negative': False, 'hh': '00',
                        'name': '+00'},
                       datetime.timedelta(hours=0), '+00'),
                      ({'negative': False, 'hh': '01',
                        'name': '+01'},
                       datetime.timedelta(hours=1), '+01'),
                      ({'negative': True, 'hh': '01',
                        'name': '-01'},
                       -datetime.timedelta(hours=1), '-01'),
                      ({'negative': False, 'hh': '12',
                        'name': '+12'},
                       datetime.timedelta(hours=12), '+12'),
                      ({'negative': True, 'hh': '12',
                        'name': '-12'},
                       -datetime.timedelta(hours=12), '-12'))

        for testtuple in testtuples:
            result = PythonTimeBuilder.build_timezone(**testtuple[0])
            self.assertEqual(result.utcoffset(None), testtuple[1])
            self.assertEqual(result.tzname(None), testtuple[2])

    def test_range_check_date(self):
        #0 isn't a valid year for a Python builder
        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_date(YYYY='0000')

        #Leap year
        #https://bitbucket.org/nielsenb/aniso8601/issues/14/parsing-ordinal-dates-should-only-allow
        with self.assertRaises(DayOutOfBoundsError):
            PythonTimeBuilder.build_date(YYYY='1981', DDD='366')

    def test_range_check_time(self):
        #Hour 24 can only represent midnight
        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh='24', mm='00', ss='01')

        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh='24', mm='00.1')

        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh='24', mm='01')

        with self.assertRaises(MidnightBoundsError):
            PythonTimeBuilder.build_time(hh='24.1')

    def test_range_check_duration(self):
        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_duration(PnY=str((datetime.timedelta.max.days // 365) + 1))

        with self.assertRaises(MonthOutOfBoundsError):
            PythonTimeBuilder.build_duration(PnM=str((datetime.timedelta.max.days // 30) + 1))

        with self.assertRaises(DayOutOfBoundsError):
            PythonTimeBuilder.build_duration(PnD=str(datetime.timedelta.max.days + 1))

        with self.assertRaises(WeekOutOfBoundsError):
            PythonTimeBuilder.build_duration(PnW=str((datetime.timedelta.max.days // 7) + 1))

        with self.assertRaises(HoursOutOfBoundsError):
            PythonTimeBuilder.build_duration(TnH=str((datetime.timedelta.max.days * 24) + 24))

        with self.assertRaises(MinutesOutOfBoundsError):
            PythonTimeBuilder.build_duration(TnM=str((datetime.timedelta.max.days * 24 * 60) + 24 * 60))

        with self.assertRaises(SecondsOutOfBoundsError):
            PythonTimeBuilder.build_duration(TnS=str((datetime.timedelta.max.days * 24 * 60 * 60) + 24 * 60 * 60))

        #Split max range across all parts
        maxpart = datetime.timedelta.max.days // 7

        with self.assertRaises(DayOutOfBoundsError):
            PythonTimeBuilder.build_duration(PnY=str((maxpart // 365) + 1),
                                             PnM=str((maxpart // 30) + 1),
                                             PnD=str((maxpart + 1)),
                                             PnW=str((maxpart // 7) + 1),
                                             TnH=str((maxpart * 24) + 1),
                                             TnM=str((maxpart * 24 * 60) + 1),
                                             TnS=str((maxpart * 24* 60 * 60) + 1))

    def test_range_check_interval(self):
        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(start=DateTuple('0007', None, None,
                                                             None, None, None),
                                             duration=DurationTuple(None, None,
                                                                    None, str(datetime.timedelta.max.days),
                                                                    None, None,
                                                                    None))

        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(start=DatetimeTuple(DateTuple('0007', None, None,
                                                                           None, None, None),
                                                                 TimeTuple('1', None, None,
                                                                           None)),
                                             duration=DurationTuple(str(datetime.timedelta.max.days // 365), None,
                                                                    None, None,
                                                                    None, None,
                                                                    None))

        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(end=DateTuple('0001', None, None,
                                                           None, None, None),
                                             duration=DurationTuple('3', None,
                                                                    None, None,
                                                                    None, None,
                                                                    None))

        with self.assertRaises(YearOutOfBoundsError):
            PythonTimeBuilder.build_interval(end=DatetimeTuple(DateTuple('0001', None, None,
                                                                           None, None, None),
                                                                 TimeTuple('1', None, None,
                                                                           None)),
                                             duration=DurationTuple('2', None,
                                                                    None, None,
                                                                    None, None,
                                                                    None))

    def test_build_week_date(self):
        weekdate = PythonTimeBuilder._build_week_date(2009, 1)
        self.assertEqual(weekdate, datetime.date(year=2008, month=12, day=29))

        weekdate = PythonTimeBuilder._build_week_date(2009, 53, isoday=7)
        self.assertEqual(weekdate, datetime.date(year=2010, month=1, day=3))

    def test_build_ordinal_date(self):
        ordinaldate = PythonTimeBuilder._build_ordinal_date(1981, 95)
        self.assertEqual(ordinaldate, datetime.date(year=1981, month=4, day=5))

    def test_iso_year_start(self):
        yearstart = PythonTimeBuilder._iso_year_start(2004)
        self.assertEqual(yearstart, datetime.date(year=2003, month=12, day=29))

        yearstart = PythonTimeBuilder._iso_year_start(2010)
        self.assertEqual(yearstart, datetime.date(year=2010, month=1, day=4))

        yearstart = PythonTimeBuilder._iso_year_start(2009)
        self.assertEqual(yearstart, datetime.date(year=2008, month=12, day=29))

    def test_date_generator(self):
        startdate = datetime.date(year=2018, month=8, day=29)
        timedelta = datetime.timedelta(days=1)
        iterations = 10

        generator = PythonTimeBuilder._date_generator(startdate,
                                                      timedelta,
                                                      iterations)

        results = list(generator)

        for dateindex in compat.range(0, 10):
            self.assertEqual(results[dateindex],
                             datetime.date(year=2018, month=8, day=29)
                             + dateindex * datetime.timedelta(days=1))

    def test_date_generator_unbounded(self):
        startdate = datetime.date(year=2018, month=8, day=29)
        timedelta = datetime.timedelta(days=5)

        generator = PythonTimeBuilder._date_generator_unbounded(startdate,
                                                                timedelta)

        #Check the first 10 results
        for dateindex in compat.range(0, 10):
            self.assertEqual(next(generator),
                             datetime.date(year=2018, month=8, day=29)
                             + dateindex * datetime.timedelta(days=5))

    def test_distribute_microseconds(self):
        self.assertEqual(PythonTimeBuilder._distribute_microseconds(1, (), ()), (1,))
        self.assertEqual(PythonTimeBuilder._distribute_microseconds(11, (0,), (10,)), (1, 1))
        self.assertEqual(PythonTimeBuilder._distribute_microseconds(211, (0, 0), (100, 10)), (2, 1, 1))

        self.assertEqual(PythonTimeBuilder._distribute_microseconds(1, (), ()), (1,))
        self.assertEqual(PythonTimeBuilder._distribute_microseconds(11, (5,), (10,)), (6, 1))
        self.assertEqual(PythonTimeBuilder._distribute_microseconds(211, (10, 5), (100, 10)), (12, 6, 1))



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.builders import TupleBuilder, DateTuple, DatetimeTuple
from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.compat import is_string
from aniso8601.date import parse_date
from aniso8601.duration import parse_duration
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import IntervalResolution
from aniso8601.time import parse_datetime, parse_time

def get_interval_resolution(isointervalstr, intervaldelimiter='/',
                            datetimedelimiter='T'):
    isointervaltuple = parse_interval(isointervalstr, intervaldelimiter=intervaldelimiter,
                                      datetimedelimiter=datetimedelimiter, builder=TupleBuilder)

    return _get_interval_resolution(isointervaltuple)

def get_repeating_interval_resolution(isointervalstr, intervaldelimiter='/',
                                      datetimedelimiter='T'):
    repeatingintervaltuple = parse_repeating_interval(isointervalstr, intervaldelimiter=intervaldelimiter,
                                                     datetimedelimiter=datetimedelimiter, builder=TupleBuilder)

    return _get_interval_resolution(repeatingintervaltuple.interval)

def _get_interval_resolution(intervaltuple):
    if intervaltuple.start is not None and intervaltuple.end is not None:
        return max(_get_interval_component_resolution(intervaltuple.start), _get_interval_component_resolution(intervaltuple.end))

    if intervaltuple.start is not None and intervaltuple.duration is not None:
        return max(_get_interval_component_resolution(intervaltuple.start), _get_interval_component_resolution(intervaltuple.duration))

    return max(_get_interval_component_resolution(intervaltuple.end), _get_interval_component_resolution(intervaltuple.duration))

def _get_interval_component_resolution(componenttuple):
    if type(componenttuple) is DateTuple:
        if componenttuple.DDD is not None:
            #YYYY-DDD
            #YYYYDDD
            return IntervalResolution.Ordinal

        if componenttuple.D is not None:
            #YYYY-Www-D
            #YYYYWwwD
            return IntervalResolution.Weekday

        if componenttuple.Www is not None:
            #YYYY-Www
            #YYYYWww
            return IntervalResolution.Week

        if componenttuple.DD is not None:
            #YYYY-MM-DD
            #YYYYMMDD
            return IntervalResolution.Day

        if componenttuple.MM is not None:
            #YYYY-MM
            return IntervalResolution.Month

        #Y[YYY]
        return IntervalResolution.Year
    elif type(componenttuple) is DatetimeTuple:
        #Datetime
        if componenttuple.time.ss is not None:
            return IntervalResolution.Seconds

        if componenttuple.time.mm is not None:
            return IntervalResolution.Minutes

        return IntervalResolution.Hours

    #Duration
    if componenttuple.TnS is not None:
        return IntervalResolution.Seconds

    if componenttuple.TnM is not None:
        return IntervalResolution.Minutes

    if componenttuple.TnH is not None:
        return IntervalResolution.Hours

    if componenttuple.PnD is not None:
        return IntervalResolution.Day

    if componenttuple.PnW is not None:
        return IntervalResolution.Week

    if componenttuple.PnM is not None:
        return IntervalResolution.Month

    return IntervalResolution.Year

def parse_interval(isointervalstr, intervaldelimiter='/',
                   datetimedelimiter='T', builder=PythonTimeBuilder):
    #Given a string representing an ISO 8601 interval, return an
    #interval built by the given builder. Valid formats are:
    #
    #<start>/<end>
    #<start>/<duration>
    #<duration>/<end>
    #
    #The <start> and <end> values can represent dates, or datetimes,
    #not times.
    #
    #The format:
    #
    #<duration>
    #
    #Is expressly not supported as there is no way to provide the additional
    #required context.

    if is_string(isointervalstr) is False:
        raise ValueError('Interval must be string.')

    if len(isointervalstr) == 0:
        raise ISOFormatError('Interval string is empty.')

    if isointervalstr[0] == 'R':
        raise ISOFormatError('ISO 8601 repeating intervals must be parsed '
                             'with parse_repeating_interval.')

    intervaldelimitercount = isointervalstr.count(intervaldelimiter)

    if intervaldelimitercount == 0:
        raise ISOFormatError('Interval delimiter "{0}" is not in interval '
                              'string "{1}".'
                             .format(intervaldelimiter, isointervalstr))

    if intervaldelimitercount > 1:
        raise ISOFormatError('{0} is not a valid ISO 8601 interval'
                             .format(isointervalstr))

    return _parse_interval(isointervalstr, builder,
                           intervaldelimiter, datetimedelimiter)

def parse_repeating_interval(isointervalstr, intervaldelimiter='/',
                             datetimedelimiter='T', builder=PythonTimeBuilder):
    #Given a string representing an ISO 8601 interval repeating, return an
    #interval built by the given builder. Valid formats are:
    #
    #Rnn/<interval>
    #R/<interval>

    if not isinstance(isointervalstr, str):
        raise ValueError('Interval must be string.')

    if len(isointervalstr) == 0:
        raise ISOFormatError('Repeating interval string is empty.')

    if isointervalstr[0] != 'R':
        raise ISOFormatError('ISO 8601 repeating interval must start '
                             'with an R.')

    if intervaldelimiter not in isointervalstr:
        raise ISOFormatError('Interval delimiter "{0}" is not in interval '
                              'string "{1}".'
                             .format(intervaldelimiter, isointervalstr))

    #Parse the number of iterations
    iterationpart, intervalpart = isointervalstr.split(intervaldelimiter, 1)

    if len(iterationpart) > 1:
        R = False
        Rnn = iterationpart[1:]
    else:
        R = True
        Rnn = None

    interval = _parse_interval(intervalpart, TupleBuilder,
                               intervaldelimiter, datetimedelimiter)

    return builder.build_repeating_interval(R=R, Rnn=Rnn, interval=interval)

def _parse_interval(isointervalstr, builder, intervaldelimiter='/',
                    datetimedelimiter='T'):
    #Returns a tuple containing the start of the interval, the end of the
    #interval, and or the interval duration

    firstpart, secondpart = isointervalstr.split(intervaldelimiter)

    if len(firstpart) == 0 or len(secondpart) == 0:
        raise ISOFormatError('{0} is not a valid ISO 8601 interval'
                             .format(isointervalstr))

    if firstpart[0] == 'P':
        #<duration>/<end>
        #Notice that these are not returned 'in order' (earlier to later), this
        #is to maintain consistency with parsing <start>/<end> durations, as
        #well as making repeating interval code cleaner. Users who desire
        #durations to be in order can use the 'sorted' operator.
        duration = parse_duration(firstpart, builder=TupleBuilder)

        #We need to figure out if <end> is a date, or a datetime
        if secondpart.find(datetimedelimiter) != -1:
            #<end> is a datetime
            endtuple = parse_datetime(secondpart,
                                      delimiter=datetimedelimiter,
                                      builder=TupleBuilder)
        else:
            endtuple = parse_date(secondpart, builder=TupleBuilder)

        return builder.build_interval(end=endtuple, duration=duration)
    elif secondpart[0] == 'P':
        #<start>/<duration>
        #We need to figure out if <start> is a date, or a datetime
        duration = parse_duration(secondpart, builder=TupleBuilder)

        if firstpart.find(datetimedelimiter) != -1:
            #<start> is a datetime
            starttuple = parse_datetime(firstpart,
                                        delimiter=datetimedelimiter,
                                        builder=TupleBuilder)
        else:
            #<start> must just be a date
            starttuple = parse_date(firstpart, builder=TupleBuilder)

        return builder.build_interval(start=starttuple,
                                      duration=duration)

    #<start>/<end>
    if firstpart.find(datetimedelimiter) != -1:
        #Both parts are datetimes
        starttuple = parse_datetime(firstpart,
                                    delimiter=datetimedelimiter,
                                    builder=TupleBuilder)
    else:
        starttuple = parse_date(firstpart, builder=TupleBuilder)

    endtuple = _parse_interval_end(secondpart,
                                   starttuple,
                                   datetimedelimiter)

    return builder.build_interval(start=starttuple, end=endtuple)

def _parse_interval_end(endstr, starttuple, datetimedelimiter):
    datestr = None
    timestr = None

    monthstr = None
    daystr = None

    concise = False

    if type(starttuple) is DateTuple:
        startdatetuple = starttuple
    else:
        #Start is a datetime
        startdatetuple = starttuple.date

    if datetimedelimiter in endstr:
        datestr, timestr = endstr.split(datetimedelimiter, 1)
    elif ':' in endstr:
        timestr = endstr
    else:
        datestr = endstr

    if timestr is not None:
        endtimetuple = parse_time(timestr, builder=TupleBuilder)

    #End is just a time
    if datestr is None:
        return endtimetuple

    #Handle backwards concise representation
    if datestr.count('-') == 1:
        monthstr, daystr = datestr.split('-')
        concise = True
    elif len(datestr) <= 2:
        daystr = datestr
        concise = True
    elif len(datestr) <= 4:
        monthstr = datestr[0:2]
        daystr = datestr[2:]
        concise = True

    if concise is True:
        concisedatestr = startdatetuple.YYYY

        #Separators required because concise elements may be missing digits
        if monthstr is not None:
            concisedatestr += '-' + monthstr
        elif startdatetuple.MM is not None:
            concisedatestr += '-' + startdatetuple.MM

        concisedatestr += '-' + daystr

        enddatetuple = parse_date(concisedatestr, builder=TupleBuilder)

        #Clear unsupplied components
        if monthstr is None:
            enddatetuple = TupleBuilder.build_date(DD=enddatetuple.DD)
        else:
            #Year not provided
            enddatetuple = TupleBuilder.build_date(MM=enddatetuple.MM,
                                                   DD=enddatetuple.DD)
    else:
        enddatetuple = parse_date(datestr, builder=TupleBuilder)

    if timestr is None:
        return enddatetuple

    return TupleBuilder.build_datetime(enddatetuple, endtimetuple)



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601 import compat
from aniso8601.builders import TupleBuilder
from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.date import parse_date
from aniso8601.decimalfraction import normalize
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import DurationResolution
from aniso8601.time import parse_time

def get_duration_resolution(isodurationstr):
    #Valid string formats are:
    #
    #PnYnMnDTnHnMnS (or any reduced precision equivalent)
    #PnW
    #P<date>T<time>
    isodurationtuple = parse_duration(isodurationstr, builder=TupleBuilder)

    if isodurationtuple.TnS is not None:
        return DurationResolution.Seconds

    if isodurationtuple.TnM is not None:
        return DurationResolution.Minutes

    if isodurationtuple.TnH is not None:
        return DurationResolution.Hours

    if isodurationtuple.PnD is not None:
        return DurationResolution.Days

    if isodurationtuple.PnW is not None:
        return DurationResolution.Weeks

    if isodurationtuple.PnM is not None:
        return DurationResolution.Months

    return DurationResolution.Years

def parse_duration(isodurationstr, builder=PythonTimeBuilder):
    #Given a string representing an ISO 8601 duration, return a
    #a duration built by the given builder. Valid formats are:
    #
    #PnYnMnDTnHnMnS (or any reduced precision equivalent)
    #PnW
    #P<date>T<time>

    if compat.is_string(isodurationstr) is False:
        raise ValueError('Duration must be string.')

    if len(isodurationstr) == 0:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                .format(isodurationstr))

    if isodurationstr[0] != 'P':
        raise ISOFormatError('ISO 8601 duration must start with a P.')

    #If Y, M, D, H, S, or W are in the string,
    #assume it is a specified duration
    if _has_any_component(isodurationstr,
                          ['Y', 'M', 'D', 'H', 'S', 'W']) is True:
        parseresult = _parse_duration_prescribed(isodurationstr)
        return builder.build_duration(**parseresult)

    if isodurationstr.find('T') != -1:
        parseresult = _parse_duration_combined(isodurationstr)
        return builder.build_duration(**parseresult)

    raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                          .format(isodurationstr))

def _parse_duration_prescribed(isodurationstr):
    #durationstr can be of the form PnYnMnDTnHnMnS or PnW

    #Make sure the end character is valid
    #https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
    if isodurationstr[-1] not in ['Y', 'M', 'D', 'H', 'S', 'W']:
        raise ISOFormatError('ISO 8601 duration must end with a valid '
                             'character.')

    #Make sure only the lowest order element has decimal precision
    durationstr = normalize(isodurationstr)

    if durationstr.count('.') > 1:
        raise ISOFormatError('ISO 8601 allows only lowest order element to '
                             'have a decimal fraction.')

    seperatoridx = durationstr.find('.')

    if seperatoridx != -1:
        remaining = durationstr[seperatoridx + 1:-1]

        #There should only ever be 1 letter after a decimal if there is more
        #then one, the string is invalid
        if remaining.isdigit() is False:
                raise ISOFormatError('ISO 8601 duration must end with '
                                     'a single valid character.')

    #Do not allow W in combination with other designators
    #https://bitbucket.org/nielsenb/aniso8601/issues/2/week-designators-should-not-be-combinable
    if (durationstr.find('W') != -1
            and _has_any_component(durationstr,
                                   ['Y', 'M', 'D', 'H', 'S']) is True):
        raise ISOFormatError('ISO 8601 week designators may not be combined '
                             'with other time designators.')

    #Parse the elements of the duration
    if durationstr.find('T') == -1:
        return _parse_duration_prescribed_notime(durationstr)

    return _parse_duration_prescribed_time(durationstr)

def _parse_duration_prescribed_notime(isodurationstr):
    #durationstr can be of the form PnYnMnD or PnW

    durationstr = normalize(isodurationstr)

    yearstr = None
    monthstr = None
    daystr = None
    weekstr = None

    weekidx = durationstr.find('W')
    yearidx = durationstr.find('Y')
    monthidx = durationstr.find('M')
    dayidx = durationstr.find('D')

    if weekidx != -1:
        weekstr = durationstr[1:-1]
    elif yearidx != -1 and monthidx != -1 and dayidx != -1:
        yearstr = durationstr[1:yearidx]
        monthstr = durationstr[yearidx + 1:monthidx]
        daystr = durationstr[monthidx + 1:-1]
    elif yearidx != -1 and monthidx != -1:
        yearstr = durationstr[1:yearidx]
        monthstr = durationstr[yearidx + 1:monthidx]
    elif monthidx != -1 and dayidx != -1:
        monthstr = durationstr[1:monthidx]
        daystr = durationstr[monthidx + 1:-1]
    elif yearidx != -1:
        yearstr = durationstr[1:-1]
    elif monthidx != -1:
        monthstr = durationstr[1:-1]
    elif dayidx != -1:
        daystr = durationstr[1:-1]
    else:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                              .format(isodurationstr))

    for componentstr in [yearstr, monthstr, daystr, weekstr]:
        if componentstr is not None:
            if '.' in componentstr:
                intstr, fractionalstr = componentstr.split('.', 1)

                if intstr.isdigit() is False:
                    raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                                          .format(isodurationstr))
            else:
                if componentstr.isdigit() is False:
                    raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                                          .format(isodurationstr))

    return {'PnY': yearstr, 'PnM': monthstr, 'PnW': weekstr, 'PnD': daystr}

def _parse_duration_prescribed_time(isodurationstr):
    #durationstr can be of the form PnYnMnDTnHnMnS

    timeidx = isodurationstr.find('T')

    datestr = isodurationstr[:timeidx]
    timestr = normalize(isodurationstr[timeidx + 1:])

    hourstr = None
    minutestr = None
    secondstr = None

    houridx = timestr.find('H')
    minuteidx = timestr.find('M')
    secondidx = timestr.find('S')

    if houridx != -1 and minuteidx != -1 and secondidx != -1:
        hourstr = timestr[0:houridx]
        minutestr = timestr[houridx + 1:minuteidx]
        secondstr = timestr[minuteidx + 1:-1]
    elif houridx != -1 and minuteidx != -1:
        hourstr = timestr[0:houridx]
        minutestr = timestr[houridx + 1:minuteidx]
    elif minuteidx != -1 and secondidx != -1:
        minutestr = timestr[0:minuteidx]
        secondstr = timestr[minuteidx + 1:-1]
    elif houridx != -1:
        hourstr = timestr[0:-1]
    elif minuteidx != -1:
        minutestr = timestr[0:-1]
    elif secondidx != -1:
        secondstr = timestr[0:-1]
    else:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                              .format(isodurationstr))

    for componentstr in [hourstr, minutestr, secondstr]:
        if componentstr is not None:
            if '.' in componentstr:
                intstr, fractionalstr = componentstr.split('.', 1)

                if intstr.isdigit() is False:
                    raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                                          .format(isodurationstr))
            else:
                if componentstr.isdigit() is False:
                    raise ISOFormatError('"{0}" is not a valid ISO 8601 duration.'
                                          .format(isodurationstr))

    #Parse any date components
    durationdict = {'PnY': None, 'PnM': None, 'PnW': None, 'PnD': None}

    if len(datestr) > 1:
        durationdict = _parse_duration_prescribed_notime(datestr)

    durationdict.update({'TnH': hourstr, 'TnM': minutestr, 'TnS': secondstr})

    return durationdict

def _parse_duration_combined(durationstr):
    #Period of the form P<date>T<time>

    #Split the string in to its component parts
    datepart, timepart = durationstr[1:].split('T', 1) #We skip the 'P'

    datevalue = parse_date(datepart, builder=TupleBuilder)
    timevalue = parse_time(timepart, builder=TupleBuilder)

    return {'PnY': datevalue.YYYY, 'PnM': datevalue.MM, 'PnD': datevalue.DD,
            'TnH': timevalue.hh, 'TnM': timevalue.mm, 'TnS': timevalue.ss}

def _has_any_component(durationstr, components):
    #Given a duration string, and a list of components, returns True
    #if any of the listed components are present, False otherwise.
    #
    #For instance:
    #durationstr = 'P1Y'
    #components = ['Y', 'M']
    #
    #returns True
    #
    #durationstr = 'P1Y'
    #components = ['M', 'D']
    #
    #returns False

    for component in components:
        if durationstr.find(component) != -1:
            return True

    return False
